"""
Plugin is used to turn on/off optional modsec rules
provided by "Secure site" feature
"""
import asyncio
from logging import getLogger

from defence360agent.contracts.config import SecureSiteConfig
from defence360agent.contracts.messages import MessageType
from defence360agent.contracts.permissions import secure_site_pro_enabled
from defence360agent.contracts.plugins import expect, MessageSink
from defence360agent.secure_site.control import Signals
from defence360agent.subsys import web_server
from defence360agent.utils import nice_iterator, recurring_check
from im360.subsys.panels.base import use_modsec_lock
from im360.subsys.panels.hosting_panel import HostingPanel

logger = getLogger(__name__)


class ModsecSecureSiteChecker(MessageSink):
    """
    Secure site modsec rules are checked in the following cases:
      - on agent startup
      - on config update (SECURE_SITE.enable setting change)
      - on user permissions update
    """
    def __init__(self):
        self._loop = None
        self._task = None
        self._previous_secure_site_status = None
        self._updating_modsec_rules_required = asyncio.Event()
        Signals.updated.connect(self._on_secure_site_update)

    def _on_secure_site_update(self, user):
        self._updating_modsec_rules_required.set()

    async def create_sink(self, loop):
        self._loop = loop
        # Note: modsec rules should be checked on agent startup
        # (first ConfigUpdate message)
        self._task = self._loop.create_task(
            recurring_check(30)(
                self.apply_secure_site_modsec_rules_for_domains
            )()
        )

    async def shutdown(self):
        if self._task is not None:
            self._task.cancel()
            await self._task

    async def apply_secure_site_modsec_rules_for_domains(self):
        await self._updating_modsec_rules_required.wait()
        self._updating_modsec_rules_required.clear()
        await self._apply_secure_site_modsec_rules_for_domains()

    @use_modsec_lock
    async def _apply_secure_site_modsec_rules_for_domains(self):
        # To avoid possible loss of an user during check
        # try to apply secure site settings for all domains on each attempt
        panel = HostingPanel()
        users_domains = await panel.get_domains_per_user()
        enabled_users_domains = {}
        disabled_users_domains = {}
        if SecureSiteConfig.ENABLED:
            async for user, domains in nice_iterator(
                users_domains.items(), chunk_size=1000
            ):
                if secure_site_pro_enabled(user):
                    enabled_users_domains[user] = domains
                else:
                    disabled_users_domains[user] = domains
        else:
            logger.info("Disable secure site settings for all domains")
            disabled_users_domains = users_domains
        try:
            logger.info("Apply secure site settings")
            updated_domains = (
                await panel.apply_secure_site_modsec_rules_for_domains(
                    enabled_users_domains=enabled_users_domains,
                    disabled_users_domains=disabled_users_domains
                )
            )
            if updated_domains:
                logger.info(
                    "Secure site settings were updated for domains: %s",
                    updated_domains
                )
                await web_server.graceful_restart()
        except Exception as exc:
            logger.exception(
                "Something went wrong during applying "
                "modsec secure site rules: %s", exc
            )

    @expect(MessageType.ConfigUpdate)
    async def on_config_update(self, message: MessageType.ConfigUpdate):
        secure_site_enabled = SecureSiteConfig.ENABLED
        if self._previous_secure_site_status != secure_site_enabled:
            self._updating_modsec_rules_required.set()
            self._previous_secure_site_status = secure_site_enabled
